/*
    Sammo's Blocks - A clone of a game by Brandon Sterner by Sam Yatchmenoff
    Copyright (C) 2002 Sam Yatchmenoff

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <SDL.h>
#include <stdlib.h>
#include <iostream>
#include <string>
#include <fstream>
#include <sstream>

using namespace std;

#define Empty 0
#define Brick 1
#define Door 2
#define Block 3

// function declerations
void load_date();

class Game {
public:
	Game();
	int load_level(const string &);
	int update();
	void draw(SDL_Surface *);
private:
	bool moved;

	int w, h;
	int *tiles;
	int &tile(int x, int y);        // makes accessing tiles a bit cleaner

	bool carrying;
	int dir;									// -1 = left, 1 = right
	int sx, sy;                      // Blockdude's location

	int vx, vy;                      // position of top left corner of viewport

	SDL_Surface *img_lsammo;
	SDL_Surface *img_rsammo;
	SDL_Surface *img_block;
	SDL_Surface *img_brick;
	SDL_Surface *img_door;
};

Game::Game() {
	img_lsammo = SDL_LoadBMP("lsammo.bmp");
	img_rsammo = SDL_LoadBMP("rsammo.bmp");
	img_block = SDL_LoadBMP("block.bmp");
	img_brick = SDL_LoadBMP("brick.bmp");
	img_door = SDL_LoadBMP("door.bmp");
}

int &Game::tile(int x, int y) {
	return tiles[x + y * w];
}

int Game::load_level(const string &filename) {
	ifstream fin(filename.c_str());

	if(!fin) {
		return -1;
	}

	fin >> sx >> sy;

	fin >> vx >> vy;

	fin >> w >> h;
	tiles = new int[w * h];
	for(int i = 0; i < w * h; i++) {
		int tmp;
		fin >> tmp;
		tiles[i] = tmp;
	}

	dir = 1;
	carrying = false;
	moved = false;
}

int Game::update() {
	int oldsx, oldsy;
	oldsx = sx;
	oldsy = sy;

	SDL_Event event;
	while(SDL_PollEvent(&event)) {
		if(event.type == SDL_KEYDOWN) {
			if(event.key.keysym.sym != SDLK_ESCAPE) {
				moved = true;
			}
			switch(event.key.keysym.sym) {
			case SDLK_ESCAPE:
				if(moved == false) {
					return 1;
				} else {
					return 2;
				}
				break;
			case SDLK_RIGHT:
				if(tile(sx + 1, sy) % 2 == 0) {
					sx++;
				}
				dir = 1;
				break;
			case SDLK_LEFT:
				if(tile(sx - 1, sy) % 2 == 0) {
					sx--;
					if(carrying && tile(sx, sy - 1) % 2 == 1) {
						carrying = 0;
						tile(sx + 1, sy - 1) = Empty;
						tile(sx + 1, sy) = Block;
					}
				}
				dir = -1;
				break;
			case SDLK_UP:
				if(carrying) {
					if(tile(sx + dir, sy) % 2 == 1 && tile(sx + dir, sy - 1) % 2 == 0 && tile(sx + dir, sy - 2) == Empty) {
						sx += dir;
						sy--;
					}
				} else {
					if(tile(sx + dir, sy) != Empty && tile(sx + dir, sy - 1) % 2 == 0) {
						sx += dir;
						sy--;
					}
				}
				break; 
			case SDLK_DOWN:
				if(carrying) {
					if(tile(sx + dir, sy) == Empty) {
						tile(sx, sy - 1) = Empty;
						tile(sx + dir, sy) = Block;
						carrying = false;
						int bx = sx + dir;
						int by = sy;
						while(tile(bx, by + 1) == Empty) {
							tile(bx, by) = Empty;
							by++;
							tile(bx, by) = Block;
						}
					} else if(tile(sx + dir, sy - 1) == Empty) {
						tile(sx, sy - 1) = Empty;
						tile(sx + dir, sy - 1) = Block;
						carrying = false;
					}
				} else {
					if(tile(sx + dir, sy) == Block && tile(sx + dir, sy - 1) != Block && tile(sx, sy - 1) == Empty) {
						tile(sx + dir, sy) = Empty;
						tile(sx, sy - 1) = Block;
						carrying = true;
					}
				}
			}
		}
	}
	while(tile(sx, sy + 1) == Empty || tile(sx, sy + 1) == Door) {
		sy++;
	}
	if(carrying) {
		tile(oldsx, oldsy - 1) = Empty;
		tile(sx, sy - 1) = Block;
	}
	return tile(sx, sy) == Door ? 3 : 0; 
}

void Game::draw(SDL_Surface *screen) {
	vx = sx - 5;  // XXX: HACKZ0R!!!!
	if(vx < 0) vx = 0;
	if(vx > w - 12) vx = w - 12;

	vy = sy - 3;
	if(vy < 0) vy = 0;
   if(vy > h - 8) vy = h - 8;

	
	SDL_Rect dest;

	dest.x = 0; dest.y = 0;
	dest.w = screen->w;
	dest.h = screen->h;
	SDL_FillRect(screen, &dest, SDL_MapRGB(screen->format, 255, 255, 255));
	
	for(int y = vy; y < vy + 8 && y < h; y++) {
		for(int x = vx; x < vx + 12 && x < w; x++) {
			dest.x = 48 * (x - vx);
			dest.y = 48 * (y - vy);
			dest.w = 48;
			dest.h = 49;
			if(tile(x, y) != Empty) {
				SDL_Surface *img;
				switch(tile(x, y)) {
					case Door: img = img_door; break;
					case Block: img = img_block; break;
					case Brick: img = img_brick; break;
				}
				SDL_BlitSurface(img, NULL, screen, &dest);
			}
		}
	}

	dest.x = 48 * (sx - vx);
	dest.y = 48 * (sy - vy);

	if(dir == -1) {
		SDL_BlitSurface(img_lsammo, NULL, screen, &dest);
	} else {
		SDL_BlitSurface(img_rsammo, NULL, screen, &dest);
	}
		
	SDL_Flip(screen);
}


int main(int argc, char **argv) {
	SDL_Init(SDL_INIT_VIDEO);
	atexit(SDL_Quit);

	SDL_Surface *screen = SDL_SetVideoMode(576, 384, 8, SDL_SWSURFACE|SDL_DOUBLEBUF);

	Game game;
	
	int level;

	if(argc > 1) {
		istringstream lvlss(argv[1]);
		lvlss >> level;
	} else {
		level = 1;
	}

	ostringstream lvlname;
	lvlname << "levels/level" << level << ".bdl";
	if(game.load_level(lvlname.str()) == -1) {
		cout << "Ack!!!\n";
		return 1;
	}

	bool done = false;
	while(!done) {
		game.draw(screen);
		int status = game.update();
		switch(status) {
		case 1:
			done = true;
			break;
		case 2:
			game.load_level(lvlname.str());
			break;
		case 3:
			cout << "You win!!\n";
			level++;
			lvlname.str("");
			lvlname << "levels/level" << level << ".bdl";
			if(game.load_level(lvlname.str()) == -1) {
				done = true;
			}
			break;
		}
	}
	return 0;
}
